from django.http import StreamingHttpResponse

class SSEManager:
    _listeners = {}

    @classmethod
    def register_listener(cls, model_name, object_id, field_name):
        if model_name not in cls._listeners:
            cls._listeners[model_name] = {}
        if object_id not in cls._listeners[model_name]:
            cls._listeners[model_name][object_id] = {}
        if field_name not in cls._listeners[model_name][object_id]:
            cls._listeners[model_name][object_id][field_name] = []
        return cls._listeners[model_name][object_id][field_name]

    @classmethod
    def notify_listeners(cls, model_name, object_id, field_name, new_value):
        listeners = cls._listeners.get(model_name, {}).get(object_id, {}).get(field_name, [])
        for listener in listeners:
            listener(new_value)

    @classmethod
    def stream_updates(cls, request, app_label, model_name, object_id, field_name):
        def event_stream():
            listeners = cls.register_listener(model_name, object_id, field_name)

            # Get the initial value from the model for sending as the first update
            from django.apps import apps
            model = apps.get_model(app_label, model_name)
            try:
                instance = model.objects.get(pk=object_id)
                last_value = getattr(instance, field_name)
            except model.DoesNotExist:
                last_value = None

            # Function to send updates to the listeners
            def send_update(new_value):
                nonlocal last_value
                if new_value != last_value:
                    last_value = new_value
                    yield f"data: {new_value}\n\n"

            # Register this listener's callback
            listeners.append(lambda new_value: send_update(new_value))

            # Start by sending the current value
            yield from send_update(last_value)

            # Handle client disconnects
            try:
                while True:
                    # SSE connection will keep streaming until the client disconnects
                    if request.META.get('HTTP_CONNECTION', '').lower() == 'close':
                        break
            except GeneratorExit:
                # Client disconnected; remove the listener
                listeners.remove(lambda new_value: send_update(new_value))
                raise

        return StreamingHttpResponse(event_stream(), content_type='text/event-stream')

sse_manager = SSEManager()


from django.views.decorators.csrf import csrf_exempt
from django.utils.decorators import method_decorator
from django.http import JsonResponse
from django.views import View
from django.apps import apps

@method_decorator(csrf_exempt, name='dispatch')
class ModelAPI(View):
    def get(self, request, app_label, model_name, object_id, field_name):
        model = apps.get_model(app_label, model_name)
        try:
            instance = model.objects.get(pk=object_id)
            field_value = getattr(instance, field_name)
            data = {
                field_name: field_value
            }
            return JsonResponse(data)
        except model.DoesNotExist:
            return JsonResponse({"error": "Object not found"}, status=404)

    def post(self, request, app_label, model_name, object_id, field_name):
        model = apps.get_model(app_label, model_name)
        instance = model.objects.get(pk=object_id)
        field_value = request.POST.get(field_name)

        if field_name and hasattr(instance, field_name):
            setattr(instance, field_name, field_value)
            instance.save(update_fields=[field_name])
            return JsonResponse({field_name: getattr(instance, field_name)})
        return JsonResponse({'error': 'Field not found or invalid data'}, status=400)



from django.db.models.signals import post_save
from django.dispatch import receiver
from django.apps import apps
from .sse import SSEManager

@receiver(post_save)
def model_update_signal(sender, instance, **kwargs):
    # Iterate over the model's fields and notify listeners if they have changed
    for field in instance._meta.fields:
        field_name = field.name
        new_value = getattr(instance, field_name, None)

        if new_value is not None:
            SSEManager.notify_listeners(sender.__name__, instance.pk, field_name, new_value)



from nicegui import ui
import requests
from .config import Config 

def bind_element_to_model(element, app_label, model_name, object_id, field_name, element_id):
    host = Config.get_host()
    api_endpoint = Config.get_api_endpoint()

    # Fetch initial data from the model
    def fetch_initial_data():
        url = f'{host}{api_endpoint}/{app_label}/{model_name}/{object_id}/{field_name}'
        response = requests.get(url)
        if response.status_code == 200:
            return response.json().get(field_name, '')
        return ''

    # Update the model when the value changes in the frontend
    def update_data(value):
        url = f'{host}{api_endpoint}/{app_label}/{model_name}/{object_id}/{field_name}/'
        requests.post(url, json={field_name: value})

    # Set the element's initial value and bind the value between frontend and backend
    element.value = fetch_initial_data()

    # Bind the element's value to backend model changes
    def on_frontend_change(e):
        new_value = e.args[0]  # Extract the new value
        update_data(new_value)  # Send updated value to the backend

    # Use the appropriate event listener for this specific element
    element.on('update:model-value', on_frontend_change) 

    # Inject JavaScript to listen to SSE updates and update the element's value
    sse_url = f'{host}{api_endpoint}/sse/{app_label}/{model_name}/{object_id}/{field_name}/'
    ui.add_body_html(f"""
        <script>
            let eventSource = new EventSource("{sse_url}");

            eventSource.onmessage = function(event) {{
                const newValue = event.data;

                // Update the NiceGUI element directly via the framework
                const element = nicegui.elements['{element_id}'];
                if (element) {{
                    element.value = newValue;
                }}
            }};

            // Cleanup: Close the EventSource connection when the page is closed or navigated away from
            window.addEventListener('beforeunload', function() {{
                if (eventSource) {{
                    eventSource.close();
                }}
            }});
        </script>
    """)

    # Add ID to the element for proper access from the injected JavaScript
    element.props(f'id={element_id}')
